package conf

import (
	"encoding/hex"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"log"
	"os"
	"time"

	"github.com/govm-net/govm/wallet"
)

// TConfig config of app
type TConfig struct {
	ServerHost     string `json:"server_host,omitempty"`
	HTTPPort       int    `json:"http_port,omitempty"`
	PrivateAddress string `json:"private_address,omitempty"`
	DbAddrType     string `json:"db_addr_type,omitempty"`
	DbServerAddr   string `json:"db_server_addr,omitempty"`
	CorePackName   []byte `json:"core_pack_name,omitempty"`
	WalletAddr     []byte `json:"wallet_addr,omitempty"`
	SignPrefix     []byte `json:"sign_prefix,omitempty"`
	PrivateKey     []byte `json:"private_key,omitempty"`
	Password       string `json:"password,omitempty"`
	WalletFile     string `json:"wallet_file,omitempty"`
	SaveLog        bool   `json:"save_log,omitempty"`
	TrustedServer  string `json:"trusted_server,omitempty"`
	CheckBlock     bool   `json:"check_block,omitempty"`
	AutoRollback   bool   `json:"auto_rollback,omitempty"`
	SaveNodeInfo   bool   `json:"save_node_info,omitempty"`
	NetID          string `json:"net_id,omitempty"`
	VerifyNetData  bool   `json:"verify_net_data,omitempty"`
	RestfulLog     bool   `json:"restful_log,omitempty"`
	TLSCertFile    string `json:"tls_cert_file,omitempty"`
	TLSKeyFile     string `json:"tls_key_file,omitempty"`
}

var (
	conf TConfig
	// Version software version
	Version string = "v0.5.14"
	// BuildTime build time
	BuildTime string
	// GitHead git head
	GitHead string

	loadFinish bool
)

func init() {
	bit := 32 << (^uint(0) >> 63)
	if bit == 32 {
		log.Println("must be 64 bit system")
		os.Exit(2)
	}
}

func loadConfig() error {
	conf.CheckBlock = true
	conf.AutoRollback = true
	conf.VerifyNetData = true
	fn := "./conf/conf.json"
	if _, err := os.Stat(fn); os.IsNotExist(err) {
		data, _ := ioutil.ReadFile("./conf/conf.bak.json")
		if len(data) > 0 {
			ioutil.WriteFile(fn, data, 0666)
		}
	}
	data, err := ioutil.ReadFile(fn)
	if err != nil {
		log.Println("fail to read file,conf.json")
		return err
	}
	err = json.Unmarshal(data, &conf)
	if err != nil {
		log.Println("fail to Unmarshal configure,conf.json")
		return err
	}
	//log.Println("config info:", conf)
	conf.CorePackName, _ = hex.DecodeString("ff0102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f")

	if conf.WalletFile == "" {
		conf.WalletFile = "./conf/wallet.key"
	}
	if conf.Password == "" {
		conf.Password = "govm_pwd@2019"
	}
	if conf.TrustedServer == "" {
		conf.TrustedServer = "http://govm.net:9090"
	}
	loadFinish = true
	return nil
}

// GetConf get configure
func GetConf() TConfig {
	if !loadFinish {
		loadConfig()
	}
	return conf
}

// Reload reload config
func Reload() error {
	old := conf
	err := loadConfig()
	if err != nil {
		log.Println("fail to reload config:", err)
		conf = old
	}
	return err
}

// LoadWallet load wallet
func LoadWallet(fileName, password string) {
	w, err := wallet.LoadWallet(fileName, password)
	if err != nil {
		if _, exist := os.Stat(fileName); !os.IsNotExist(exist) {
			fmt.Println("fail to load wallet.", err)
			os.Exit(4)
		}
		os.Rename(fileName, fileName+".error")
		w, err = wallet.LoadWallet("./conf/base.dat", password)
		if err != nil {
			w.Key = wallet.NewPrivateKey()
			pubKey := wallet.GetPublicKey(w.Key, wallet.EAddrTypeDefault)
			w.Address = wallet.PublicKeyToAddress(pubKey, wallet.EAddrTypeDefault)
		} else {
			now := time.Now().Unix() * 1000
			w.Key, w.SignPrefix = wallet.NewChildPrivateKeyOfIBS(w.Key, uint64(now))
		}
		wallet.SaveWallet(fileName, password, w.Address, w.Key, w.SignPrefix)
	}
	conf.PrivateKey = w.Key
	conf.WalletAddr = w.Address
	conf.SignPrefix = w.SignPrefix
}
